@@ -23,6 +23,7 @@
 #![deny(missing_docs)]
 #![warn(noop_method_call)]
 #![deny(unreachable_pub)]
+#![deny(clippy::all)]
 #![deny(clippy::await_holding_lock)]
 #![deny(clippy::cargo_common_metadata)]
 #![deny(clippy::cast_lossless)]
@@ -44,6 +45,7 @@
 #![warn(clippy::option_option)]
 #![warn(clippy::rc_buffer)]
 #![deny(clippy::ref_option_ref)]
+#![warn(clippy::semicolon_if_nothing_returned)]
 #![warn(clippy::trait_duplication_in_bounds)]
 #![deny(clippy::unnecessary_wraps)]
 #![warn(clippy::unseparated_literal_suffix)]
@@ -54,11 +56,10 @@ use tor_netdir::{fallback::FallbackDir, NetDir};
 use tor_proto::circuit::{CircParameters, ClientCirc, UniqId};
 use tor_rtcompat::Runtime;
 
-use futures::task::SpawnExt;
 use std::convert::TryInto;
-use std::sync::{Arc, Weak};
-use std::time::Duration;
-use tracing::{debug, warn};
+use std::sync::{Arc, Mutex};
+use std::time::Instant;
+use tracing::{debug, info, warn};
 
 pub mod build;
 mod config;
@@ -66,6 +67,7 @@ mod err;
 mod impls;
 mod mgr;
 pub mod path;
+mod preemptive;
 mod timeouts;
 mod usage;
 
@@ -74,9 +76,10 @@ pub use usage::{IsolationToken, StreamIsolation, StreamIsolationBuilder, TargetP
 
 pub use config::{
     CircMgrConfig, CircMgrConfigBuilder, CircuitTiming, CircuitTimingBuilder, PathConfig,
-    PathConfigBuilder, RequestTiming, RequestTimingBuilder,
+    PathConfigBuilder, PreemptiveCircuitConfig, PreemptiveCircuitConfigBuilder,
 };
 
+use crate::preemptive::PreemptiveCircuitPredictor;
 use usage::TargetCircUsage;
 
 /// A Result type as returned from this crate.
@@ -149,6 +152,8 @@ impl<'a> DirInfo<'a> {
 pub struct CircMgr<R: Runtime> {
     /// The underlying circuit manager object that implements our behavior.
     mgr: Arc<mgr::AbstractCircMgr<build::CircuitBuilder<R>, R>>,
+    /// A preemptive circuit predictor, for, uh, building circuits preemptively.
+    predictor: Arc<Mutex<PreemptiveCircuitPredictor>>,
 }
 
 impl<R: Runtime> CircMgr<R> {
@@ -163,11 +168,15 @@ impl<R: Runtime> CircMgr<R> {
         SM: tor_persist::StateMgr + Send + Sync + 'static,
     {
         let CircMgrConfig {
-            path_config,
-            request_timing,
+            path_rules,
             circuit_timing,
+            preemptive_circuits,
         } = config;
 
+        let preemptive = Arc::new(Mutex::new(PreemptiveCircuitPredictor::new(
+            preemptive_circuits,
+        )));
+
         let guardmgr = tor_guardmgr::GuardMgr::new(runtime.clone(), storage.clone())?;
 
         let storage_handle = storage.create_handle(PARETO_TIMEOUT_DATA_KEY);
@@ -175,22 +184,59 @@ impl<R: Runtime> CircMgr<R> {
         let builder = build::CircuitBuilder::new(
             runtime.clone(),
             chanmgr,
-            path_config,
+            path_rules,
             storage_handle,
             guardmgr,
         );
-        let mgr =
-            mgr::AbstractCircMgr::new(builder, runtime.clone(), request_timing, circuit_timing);
-        let circmgr = Arc::new(CircMgr { mgr: Arc::new(mgr) });
-
-        runtime.spawn(continually_expire_circuits(
-            runtime.clone(),
-            Arc::downgrade(&circmgr),
-        ))?;
+        let mgr = mgr::AbstractCircMgr::new(builder, runtime.clone(), circuit_timing);
+        let circmgr = Arc::new(CircMgr {
+            mgr: Arc::new(mgr),
+            predictor: preemptive,
+        });
 
         Ok(circmgr)
     }
 
+    /// Try to change our configuration settings to `new_config`.
+    ///
+    /// The actual behavior here will depend on the value of `how`.
+    pub fn reconfigure(
+        &self,
+        new_config: &CircMgrConfig,
+        how: tor_config::Reconfigure,
+    ) -> std::result::Result<(), tor_config::ReconfigureError> {
+        let old_path_rules = self.mgr.peek_builder().path_config();
+        let predictor = self.predictor.lock().expect("poisoned lock");
+        let preemptive_circuits = predictor.config();
+        if preemptive_circuits.initial_predicted_ports
+            != new_config.preemptive_circuits.initial_predicted_ports
+        {
+            // This change has no effect, since the list of ports was _initial_.
+            how.cannot_change("preemptive_circuits.initial_predicted_ports")?;
+        }
+
+        if how == tor_config::Reconfigure::CheckAllOrNothing {
+            return Ok(());
+        }
+
+        let discard_circuits = !new_config.path_rules.more_permissive_than(&old_path_rules);
+
+        self.mgr
+            .peek_builder()
+            .set_path_config(new_config.path_rules.clone());
+        self.mgr
+            .set_circuit_timing(new_config.circuit_timing.clone());
+        predictor.set_config(new_config.preemptive_circuits.clone());
+
+        if discard_circuits {
+            // TODO(nickm): Someday, we might want to take a more lenient approach, and only
+            // retire those circuits that do not conform to the new path rules.
+            info!("Path configuration has become more restrictive: retiring existing circuits.");
+            self.retire_all_circuits();
+        }
+        Ok(())
+    }
+
     /// Reload state from the state manager.
     ///
     /// We only call this method if we _don't_ have the lock on the state
@@ -224,6 +270,18 @@ impl<R: Runtime> CircMgr<R> {
         self.mgr.peek_builder().update_network_parameters(p);
     }
 
+    /// Return true if `netdir` has enough information to be used for this
+    /// circuit manager.
+    ///
+    /// (This will check whether the netdir is missing any primary guard
+    /// microdescriptors)
+    pub fn netdir_is_sufficient(&self, netdir: &NetDir) -> bool {
+        self.mgr
+            .peek_builder()
+            .guardmgr()
+            .netdir_is_sufficient(netdir)
+    }
+
     /// Reconfigure this circuit manager using the latest network directory.
     ///
     /// This should be called on _any_ change to the network, as opposed to
@@ -235,7 +293,7 @@ impl<R: Runtime> CircMgr<R> {
 
     /// Return a circuit suitable for sending one-hop BEGINDIR streams,
     /// launching it if necessary.
-    pub async fn get_or_launch_dir(&self, netdir: DirInfo<'_>) -> Result<Arc<ClientCirc>> {
+    pub async fn get_or_launch_dir(&self, netdir: DirInfo<'_>) -> Result<ClientCirc> {
         self.expire_circuits();
         let usage = TargetCircUsage::Dir;
         self.mgr.get_or_launch(&usage, netdir).await
@@ -251,26 +309,84 @@ impl<R: Runtime> CircMgr<R> {
         netdir: DirInfo<'_>, // TODO: This has to be a NetDir.
         ports: &[TargetPort],
         isolation: StreamIsolation,
-    ) -> Result<Arc<ClientCirc>> {
+    ) -> Result<ClientCirc> {
         self.expire_circuits();
+        let time = Instant::now();
+        {
+            let mut predictive = self.predictor.lock().expect("preemptive lock poisoned");
+            if ports.is_empty() {
+                predictive.note_usage(None, time);
+            } else {
+                for port in ports.iter() {
+                    predictive.note_usage(Some(*port), time);
+                }
+            }
+        }
         let ports = ports.iter().map(Clone::clone).collect();
         let usage = TargetCircUsage::Exit { ports, isolation };
         self.mgr.get_or_launch(&usage, netdir).await
     }
 
+    /// Launch circuits preemptively, using the preemptive circuit predictor's predictions.
+    ///
+    /// # Note
+    ///
+    /// This function is invoked periodically from the
+    /// `arti-client` crate, based on timings from the network
+    /// parameters. As with `launch_timeout_testing_circuit_if_appropriate`, this
+    /// should ideally be refactored to be internal to this crate, and not be a
+    /// public API here.
+    pub async fn launch_circuits_preemptively(&self, netdir: DirInfo<'_>) {
+        debug!("Checking preemptive circuit predictions.");
+        let (circs, threshold) = {
+            let preemptive = self.predictor.lock().expect("preemptive lock poisoned");
+            let threshold = preemptive.config().disable_at_threshold;
+            (preemptive.predict(), threshold)
+        };
+
+        if self.mgr.n_circs() >= threshold {
+            return;
+        }
+
+        let futures = circs
+            .iter()
+            .map(|usage| self.mgr.get_or_launch(usage, netdir));
+        let results = futures::future::join_all(futures).await;
+        for (i, result) in results.iter().enumerate() {
+            match result {
+                Ok(_) => debug!("Circuit exists (or was created) for {:?}", circs[i]),
+                Err(e) => warn!("Failed to build preemptive circuit {:?}: {}", circs[i], e),
+            }
+        }
+    }
+
     /// If `circ_id` is the unique identifier for a circuit that we're
     /// keeping track of, don't give it out for any future requests.
     pub fn retire_circ(&self, circ_id: &UniqId) {
         let _ = self.mgr.take_circ(circ_id);
     }
 
+    /// Mark every circuit that we have launched so far as unsuitable for
+    /// any future requests.  This won't close existing circuits that have
+    /// streams attached to them, but it will prevent any future streams from
+    /// being attached.
+    ///
+    /// TODO: we may want to expose this eventually.  If we do, we should
+    /// be very clear that you don't want to use it haphazardly.
+    pub(crate) fn retire_all_circuits(&self) {
+        self.mgr.retire_all_circuits();
+    }
+
     /// Expire every circuit that has been dirty for too long.
     ///
     /// Expired circuits are not closed while they still have users,
     /// but they are no longer given out for new requests.
     fn expire_circuits(&self) {
-        // TODO: I would prefer not to call this at every request, but it
-        // should be fine for now.
+        // TODO: I would prefer not to call this at every request, but
+        // it should be fine for now.  (At some point we may no longer
+        // need this, or might not need to call it so often, now that
+        // our circuit expiration runs on scheduld timers via
+        // spawn_expiration_task.)
         let now = self.mgr.peek_runtime().now();
         self.mgr.expire_circs(now);
     }
@@ -315,29 +431,6 @@ impl<R: Runtime> CircMgr<R> {
     }
 }
 
-/// Periodically expire any circuits that should no longer be given
-/// out for requests.
-///
-/// Exit when we find that `circmgr` is dropped.
-///
-/// This is a daemon task: it runs indefinitely in the background.
-async fn continually_expire_circuits<R: Runtime>(runtime: R, circmgr: Weak<CircMgr<R>>) {
-    // TODO: This is too long for accuracy and too short for
-    // efficiency.  Instead we should have a more clever scheduling
-    // algorithm somehow that gets updated when we have new or newly
-    // dirty circuits only.
-    let interval = Duration::from_secs(5);
-
-    loop {
-        runtime.sleep(interval).await;
-        if let Some(cm) = Weak::upgrade(&circmgr) {
-            cm.expire_circuits();
-        } else {
-            break;
-        }
-    }
-}
-
 #[cfg(test)]
 mod test {
     #![allow(clippy::unwrap_used)]
