@@ -1,8 +1,10 @@
 //! Code related to tracking what activities a circuit can be used for.
 
 use rand::Rng;
+use serde::{Deserialize, Serialize};
 use std::sync::atomic::{AtomicU64, Ordering};
 use std::sync::Arc;
+use tracing::debug;
 
 use crate::path::{dirpath::DirPathBuilder, exitpath::ExitPathBuilder, TorPath};
 use tor_guardmgr::{GuardMgr, GuardMonitor, GuardUsable};
@@ -10,6 +12,7 @@ use tor_netdir::Relay;
 use tor_netdoc::types::policy::PortPolicy;
 use tor_rtcompat::Runtime;
 
+use crate::mgr::{abstract_spec_find_supported, AbstractCirc, OpenEntry};
 use crate::{Error, Result};
 
 /// An exit policy, as supported by the last hop of a circuit.
@@ -25,7 +28,7 @@ pub(crate) struct ExitPolicy {
 ///
 /// Ordinarily, this is a TCP port, plus a flag to indicate whether we
 /// must support IPv4 or IPv6.
-#[derive(Clone, Copy, Debug, Eq, PartialEq)]
+#[derive(Clone, Copy, Debug, Deserialize, Eq, PartialEq, Hash, PartialOrd, Ord, Serialize)]
 pub struct TargetPort {
     /// True if this is a request to connect to an IPv6 address
     ipv6: bool,
@@ -104,7 +107,7 @@ impl TargetPort {
 //
 // This type is re-exported by `arti-client`: any changes to it must be
 // reflected in `arti-client`'s version.
-#[derive(Clone, Copy, Debug, PartialEq, Eq)]
+#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
 pub struct IsolationToken(u64);
 
 #[allow(clippy::new_without_default)]
@@ -141,7 +144,7 @@ impl IsolationToken {
 ///
 /// If two streams are isolated from one another, they may not share
 /// a circuit.
-#[derive(Copy, Clone, Eq, Debug, PartialEq, derive_builder::Builder)]
+#[derive(Copy, Clone, Eq, Debug, PartialEq, PartialOrd, Ord, derive_builder::Builder)]
 pub struct StreamIsolation {
     /// Any isolation token set on the stream.
     #[builder(default = "IsolationToken::no_isolation()")]
@@ -165,6 +168,12 @@ impl StreamIsolation {
     pub fn builder() -> StreamIsolationBuilder {
         StreamIsolationBuilder::new()
     }
+
+    /// Return true if this StreamIsolation can share a circuit with
+    /// `other`.
+    fn may_share_circuit(&self, other: &StreamIsolation) -> bool {
+        self == other
+    }
 }
 
 impl StreamIsolationBuilder {
@@ -199,7 +208,7 @@ impl ExitPolicy {
 ///
 /// This type should stay internal to the circmgr crate for now: we'll probably
 /// want to refactor it a lot.
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
 pub(crate) enum TargetCircUsage {
     /// Use for BEGINDIR-based non-anonymous directory connections
     Dir,
@@ -215,6 +224,21 @@ pub(crate) enum TargetCircUsage {
     },
     /// For a circuit is only used for the purpose of building it.
     TimeoutTesting,
+    /// For internal usage only: build a circuit preemptively, to reduce wait times.
+    ///
+    /// # Warning
+    ///
+    /// This **MUST NOT** be used by code outside of the preemptive circuit predictor. In
+    /// particular, this usage doesn't support stream isolation, so using it to ask for
+    /// circuits (for example, by passing it to `get_or_launch`) could be unsafe!
+    Preemptive {
+        /// A port the circuit has to allow, if specified.
+        ///
+        /// If this is `None`, we just want a circuit capable of doing DNS resolution.
+        port: Option<TargetPort>,
+        /// The number of exit circuits needed for a port
+        circs: usize,
+    },
 }
 
 /// The purposes for which a circuit is usable.
@@ -223,7 +247,7 @@ pub(crate) enum TargetCircUsage {
 /// want to refactor it a lot.
 #[derive(Clone, Debug, PartialEq)]
 pub(crate) enum SupportedCircUsage {
-    /// Useable for BEGINDIR-based non-anonymous directory connections
+    /// Usable for BEGINDIR-based non-anonymous directory connections
     Dir,
     /// Usable to exit to a set of ports.
     Exit {
@@ -257,6 +281,23 @@ impl TargetCircUsage {
                 let (path, mon, usable) = DirPathBuilder::new().pick_path(rng, netdir, guards)?;
                 Ok((path, SupportedCircUsage::Dir, mon, usable))
             }
+            TargetCircUsage::Preemptive { port, .. } => {
+                // FIXME(eta): this is copypasta from `TargetCircUsage::Exit`.
+                let (path, mon, usable) = ExitPathBuilder::from_target_ports(port.iter().copied())
+                    .pick_path(rng, netdir, guards, config)?;
+                let policy = path
+                    .exit_policy()
+                    .expect("ExitPathBuilder gave us a one-hop circuit?");
+                Ok((
+                    path,
+                    SupportedCircUsage::Exit {
+                        policy,
+                        isolation: None,
+                    },
+                    mon,
+                    usable,
+                ))
+            }
             TargetCircUsage::Exit {
                 ports: p,
                 isolation,
@@ -311,9 +352,21 @@ impl crate::mgr::AbstractSpec for SupportedCircUsage {
                     isolation: i2,
                 },
             ) => {
-                i1.map(|i1| i1 == *i2).unwrap_or(true)
+                i1.map(|i1| i1.may_share_circuit(i2)).unwrap_or(true)
                     && p2.iter().all(|port| p1.allows_port(*port))
             }
+            (Exit { policy, isolation }, TargetCircUsage::Preemptive { port, .. }) => {
+                if isolation.is_some() {
+                    // If the circuit has a stream isolation token, we might not be able to use it
+                    // for new streams that don't share it.
+                    return false;
+                }
+                if let Some(p) = port {
+                    policy.allows_port(*p)
+                } else {
+                    true
+                }
+            }
             (Exit { .. } | NoUsage, TargetCircUsage::TimeoutTesting) => true,
             (_, _) => false,
         }
@@ -324,13 +377,19 @@ impl crate::mgr::AbstractSpec for SupportedCircUsage {
 
         match (self, usage) {
             (Dir, TargetCircUsage::Dir) => Ok(()),
+            // This usage is only used to create circuits preemptively, and doesn't actually
+            // correspond to any streams; accordingly, we don't need to modify the circuit's
+            // acceptable usage at all.
+            (Exit { .. }, TargetCircUsage::Preemptive { .. }) => Ok(()),
             (
                 Exit {
                     isolation: ref mut i1,
                     ..
                 },
                 TargetCircUsage::Exit { isolation: i2, .. },
-            ) if i1.map(|i1| i1 == *i2).unwrap_or(true) => {
+            ) if i1.map(|i1| i1.may_share_circuit(i2)).unwrap_or(true) => {
+                // Once we have more complex isolation, this assignment
+                // won't be correct.
                 *i1 = Some(*i2);
                 Ok(())
             }
@@ -341,36 +400,75 @@ impl crate::mgr::AbstractSpec for SupportedCircUsage {
             (_, _) => Err(Error::UsageNotSupported("Incompatible usage".into())),
         }
     }
+
+    fn find_supported<'a, 'b, C: AbstractCirc>(
+        list: impl Iterator<Item = &'b mut OpenEntry<Self, C>>,
+        usage: &TargetCircUsage,
+    ) -> Vec<&'b mut OpenEntry<Self, C>> {
+        match usage {
+            TargetCircUsage::Preemptive { circs, .. } => {
+                let supported = abstract_spec_find_supported(list, usage);
+                // We need to have at least two circuits that support `port` in order
+                // to reuse them; otherwise, we must create a new circuit, so
+                // that we get closer to having two circuits.
+                debug!(
+                    "preemptive usage {:?} matches {} active circuits",
+                    usage,
+                    supported.len()
+                );
+                if supported.len() >= *circs {
+                    supported
+                } else {
+                    vec![]
+                }
+            }
+            _ => abstract_spec_find_supported(list, usage),
+        }
+    }
 }
 
 #[cfg(test)]
 mod test {
     #![allow(clippy::unwrap_used)]
     use super::*;
+    use crate::path::OwnedPath;
     use crate::test::OptDummyGuardMgr;
+    use std::convert::TryFrom;
+    use tor_linkspec::ChanTarget;
     use tor_netdir::testnet;
 
     #[test]
     fn exit_policy() {
-        let network = testnet::construct_netdir()
-            .unwrap()
-            .unwrap_if_sufficient()
-            .unwrap();
+        use tor_netdir::testnet::construct_custom_netdir;
+        use tor_netdoc::doc::netstatus::RelayFlags;
+
+        let network = construct_custom_netdir(|idx, nb| {
+            if (0x21..0x27).contains(&idx) {
+                nb.rs.add_flags(RelayFlags::BAD_EXIT);
+            }
+        })
+        .unwrap()
+        .unwrap_if_sufficient()
+        .unwrap();
 
         // Nodes with ID 0x0a through 0x13 and 0x1e through 0x27 are
         // exits.  Odd-numbered ones allow only ports 80 and 443;
-        // even-numbered ones allow all ports.
+        // even-numbered ones allow all ports.  Nodes with ID 0x21
+        // through 0x27 are bad exits.
         let id_noexit = [0x05; 32].into();
         let id_webexit = [0x11; 32].into();
         let id_fullexit = [0x20; 32].into();
+        let id_badexit = [0x25; 32].into();
 
         let not_exit = network.by_id(&id_noexit).unwrap();
         let web_exit = network.by_id(&id_webexit).unwrap();
         let full_exit = network.by_id(&id_fullexit).unwrap();
+        let bad_exit = network.by_id(&id_badexit).unwrap();
 
         let ep_none = ExitPolicy::from_relay(&not_exit);
         let ep_web = ExitPolicy::from_relay(&web_exit);
         let ep_full = ExitPolicy::from_relay(&full_exit);
+        let ep_bad = ExitPolicy::from_relay(&bad_exit);
 
         assert!(!ep_none.allows_port(TargetPort::ipv4(80)));
         assert!(!ep_none.allows_port(TargetPort::ipv4(9999)));
@@ -383,15 +481,18 @@ mod test {
         assert!(ep_full.allows_port(TargetPort::ipv4(443)));
         assert!(ep_full.allows_port(TargetPort::ipv4(9999)));
 
+        assert!(!ep_bad.allows_port(TargetPort::ipv4(80)));
+
         // Note that nobody in the testdir::network allows ipv6.
         assert!(!ep_none.allows_port(TargetPort::ipv6(80)));
         assert!(!ep_web.allows_port(TargetPort::ipv6(80)));
         assert!(!ep_full.allows_port(TargetPort::ipv6(80)));
+        assert!(!ep_bad.allows_port(TargetPort::ipv6(80)));
 
         // Check is_supported_by while we're here.
-        // TODO: Make sure that if BadExit is set, this function returns no
         assert!(TargetPort::ipv4(80).is_supported_by(&web_exit));
         assert!(!TargetPort::ipv6(80).is_supported_by(&web_exit));
+        assert!(!TargetPort::ipv6(80).is_supported_by(&bad_exit));
     }
 
     #[test]
@@ -428,6 +529,8 @@ mod test {
             policy,
             isolation: None,
         };
+        let supp_none = SupportedCircUsage::NoUsage;
+
         let targ_80_v4 = TargetCircUsage::Exit {
             ports: vec![TargetPort::ipv4(80)],
             isolation,
@@ -448,6 +551,7 @@ mod test {
             ports: vec![TargetPort::ipv6(999)],
             isolation,
         };
+        let targ_testing = TargetCircUsage::TimeoutTesting;
 
         assert!(supp_dir.supports(&targ_dir));
         assert!(!supp_dir.supports(&targ_80_v4));
@@ -462,6 +566,14 @@ mod test {
         assert!(supp_exit_no_iso.supports(&targ_80_v4));
         assert!(supp_exit_no_iso.supports(&targ_80_v4_iso2));
         assert!(!supp_exit_no_iso.supports(&targ_80_23_v4));
+        assert!(!supp_none.supports(&targ_dir));
+        assert!(!supp_none.supports(&targ_80_23_v4));
+        assert!(!supp_none.supports(&targ_80_v4_iso2));
+        assert!(!supp_dir.supports(&targ_testing));
+        assert!(supp_exit.supports(&targ_testing));
+        assert!(supp_exit_no_iso.supports(&targ_testing));
+        assert!(supp_exit_iso2.supports(&targ_testing));
+        assert!(supp_none.supports(&targ_testing));
     }
 
     #[test]
@@ -498,6 +610,7 @@ mod test {
             policy,
             isolation: None,
         };
+        let supp_none = SupportedCircUsage::NoUsage;
         let targ_exit = TargetCircUsage::Exit {
             ports: vec![TargetPort::ipv4(80)],
             isolation,
@@ -506,10 +619,12 @@ mod test {
             ports: vec![TargetPort::ipv4(80)],
             isolation: isolation2,
         };
+        let targ_testing = TargetCircUsage::TimeoutTesting;
 
         // not allowed, do nothing
         let mut supp_dir_c = supp_dir.clone();
         assert!(supp_dir_c.restrict_mut(&targ_exit).is_err());
+        assert!(supp_dir_c.restrict_mut(&targ_testing).is_err());
         assert_eq!(supp_dir, supp_dir_c);
 
         let mut supp_exit_c = supp_exit.clone();
@@ -524,6 +639,11 @@ mod test {
         assert!(supp_exit_iso2_c.restrict_mut(&targ_exit).is_err());
         assert_eq!(supp_exit_iso2, supp_exit_iso2_c);
 
+        let mut supp_none_c = supp_none.clone();
+        assert!(supp_none_c.restrict_mut(&targ_exit).is_err());
+        assert!(supp_none_c.restrict_mut(&targ_dir).is_err());
+        assert_eq!(supp_none_c, supp_none);
+
         // allowed but nothing to do
         let mut supp_dir_c = supp_dir.clone();
         supp_dir_c.restrict_mut(&targ_dir).unwrap();
@@ -535,8 +655,13 @@ mod test {
 
         let mut supp_exit_iso2_c = supp_exit_iso2.clone();
         supp_exit_iso2_c.restrict_mut(&targ_exit_iso2).unwrap();
+        supp_none_c.restrict_mut(&targ_testing).unwrap();
         assert_eq!(supp_exit_iso2, supp_exit_iso2_c);
 
+        let mut supp_none_c = supp_none.clone();
+        supp_none_c.restrict_mut(&targ_testing).unwrap();
+        assert_eq!(supp_none_c, supp_none);
+
         // allowed, do something
         let mut supp_exit_no_iso_c = supp_exit_no_iso.clone();
         supp_exit_no_iso_c.restrict_mut(&targ_exit).unwrap();
@@ -564,12 +689,15 @@ mod test {
         // Only doing basic tests for now.  We'll test the path
         // building code a lot more closely in the tests for TorPath
         // and friends.
+
+        // First, a one-hop directory circuit
         let (p_dir, u_dir, _, _) = TargetCircUsage::Dir
             .build_path(&mut rng, di, guards, &config)
             .unwrap();
         assert!(matches!(u_dir, SupportedCircUsage::Dir));
         assert_eq!(p_dir.len(), 1);
 
+        // Now an exit circuit, to port 995.
         let tok1 = IsolationToken::new();
         let isolation = StreamIsolationBuilder::new()
             .owner_token(tok1)
@@ -592,5 +720,75 @@ mod test {
         ));
         assert!(u_exit.supports(&exit_usage));
         assert_eq!(p_exit.len(), 3);
+
+        // Now try testing circuits.
+        let (path, usage, _, _) = TargetCircUsage::TimeoutTesting
+            .build_path(&mut rng, di, guards, &config)
+            .unwrap();
+        let path = match OwnedPath::try_from(&path).unwrap() {
+            OwnedPath::ChannelOnly(_) => panic!("Impossible path type."),
+            OwnedPath::Normal(p) => p,
+        };
+        assert_eq!(path.len(), 3);
+
+        // Make sure that the usage is correct.
+        let last_relay = netdir.by_id(path[2].ed_identity()).unwrap();
+        let policy = ExitPolicy::from_relay(&last_relay);
+        // We'll always get exits for these, since we try to build
+        // paths with an exit if there are any exits.
+        assert!(policy.allows_some_port());
+        assert!(last_relay.policies_allow_some_port());
+        assert_eq!(
+            usage,
+            SupportedCircUsage::Exit {
+                policy,
+                isolation: None
+            }
+        );
+    }
+
+    #[test]
+    fn build_testing_noexit() {
+        // Here we'll try to build paths for testing circuits on a network
+        // with no exits.
+        let mut rng = rand::thread_rng();
+        let netdir = testnet::construct_custom_netdir(|_idx, bld| {
+            bld.md.parse_ipv4_policy("reject 1-65535").unwrap();
+        })
+        .unwrap()
+        .unwrap_if_sufficient()
+        .unwrap();
+        let di = (&netdir).into();
+        let config = crate::PathConfig::default();
+        let guards: OptDummyGuardMgr<'_> = None;
+
+        let (path, usage, _, _) = TargetCircUsage::TimeoutTesting
+            .build_path(&mut rng, di, guards, &config)
+            .unwrap();
+        assert_eq!(path.len(), 3);
+        assert_eq!(usage, SupportedCircUsage::NoUsage);
+    }
+
+    #[test]
+    fn build_isolation() {
+        let no_isolation = StreamIsolation::no_isolation();
+        let no_isolation2 = StreamIsolation::builder()
+            .owner_token(IsolationToken::no_isolation())
+            .stream_token(IsolationToken::no_isolation())
+            .build()
+            .unwrap();
+        assert_eq!(no_isolation, no_isolation2);
+        assert!(no_isolation.may_share_circuit(&no_isolation2));
+
+        let tok = IsolationToken::new();
+        let some_isolation = StreamIsolation::builder().owner_token(tok).build().unwrap();
+        let some_isolation2 = StreamIsolation::builder()
+            .stream_token(tok)
+            .build()
+            .unwrap();
+        assert!(!no_isolation.may_share_circuit(&some_isolation));
+        assert!(!no_isolation.may_share_circuit(&some_isolation2));
+        assert!(!some_isolation.may_share_circuit(&some_isolation2));
+        assert!(some_isolation.may_share_circuit(&some_isolation));
     }
 }
